require("defines")
require("util")
-- debug_status = 1
debug_mod_name = "Megamind"
debug_file = debug_mod_name .. "-debug.txt"
require("utils")
require("config")

local status_created = 0
local status_charging = 1
local status_scanning = 2
local status_emitting = 3

--------------------------------------------------------------------------------------
function init_globals()
	global.tick = global.tick or 0
	global.controllers = global.controllers or {}
	global.released = global.released or {} -- released aliens
end

--------------------------------------------------------------------------------------
script.on_init( 
	function()
		debug_print( "on_init" )
		init_globals()
	end
)

--------------------------------------------------------------------------------------
function on_configuration_changed(data)
	-- detect any mod or game version change
	if data.mod_changes ~= nil then
		local changes = data.mod_changes[debug_mod_name]
		if changes ~= nil then
			debug_print( "update mod: ", debug_mod_name, " ", tostring(changes.old_version), " to ", tostring(changes.new_version) )
		
			init_globals()
			
			for _,force in pairs(game.forces) do
				force.reset_recipes()
				force.reset_technologies()
			end
		end
	end
end

script.on_configuration_changed(on_configuration_changed)

--------------------------------------------------------------------------------------
function on_creation( event )
	local ent = event.created_entity
	local player_index = event.player_index
	
	if ent.name == "mind-controller" then
		debug_print( "mind-controller created" )
		
		local surface = ent.surface
		local force = ent.force
		local position = ent.position
		
		ent.destroy()
		ent = surface.create_entity({name = "mind-controller-invisible", force = force, position = position})
		local charging = surface.create_entity({name = "mind-controller-charging", force = game.forces.neutral, position = position})

		table.insert( global.controllers,
			{
				entity_invisible = ent,
				entity_visible = charging,
				status = status_created,
				surface = ent.surface,
				force = ent.force,
				position = ent.position,
				inv_input = ent.get_inventory(defines.inventory.assembling_machine_input),
				inv_output = ent.get_inventory(defines.inventory.assembling_machine_output),
				friends = {},
			}
		)

		debug_print( "controllers ", #global.controllers )
	end
end

script.on_event(defines.events.on_built_entity, on_creation )
script.on_event(defines.events.on_robot_built_entity, on_creation )

--------------------------------------------------------------------------------------
function on_destruction( event )
	local ent = event.entity
	
	if ent.name == "mind-controller-invisible" or ent.name == "mind-controller" then
		debug_print( "mind-controller destroyed" )
		
		for k, controller in pairs(global.controllers) do	
			if controller.entity_invisible == ent then
				controller.inv_input.clear()
				controller.inv_output.clear()
				if controller.entity_visible and controller.entity_visible.valid then
					controller.entity_visible.destroy()
				end
				
				for _, friend in pairs( controller.friends ) do
					if friend and friend.valid then
						release_friend( controller, friend )
					end
				end
				
				table.remove(global.controllers,k)
				
				debug_print( "controllers " , #global.controllers )
				break
			end
		end
	end
	
end

script.on_event(defines.events.on_entity_died, on_destruction )
script.on_event(defines.events.on_robot_pre_mined, on_destruction )
script.on_event(defines.events.on_preplayer_mined_item, on_destruction )

--------------------------------------------------------------------------------------
function on_tick(event)
	local ent
	
	if global.tick <= 0 then
		global.tick = 55
		
		for _, controller in pairs(global.controllers) do	
			if controller.status == status_created then
				debug_print( "charging!" )
				controller.status = status_charging

				ent = controller.entity_visible
				if ent then ent.destroy() end
				-- ent = controller.surface.create_entity({name = "mind-controller-charging", force = game.forces.neutral, position = pos_offset(controller.position,{x=2,y=0})})
				ent = controller.surface.create_entity({name = "mind-controller-charging", force = game.forces.neutral, position = controller.position})
				ent.operable = false
				ent.get_inventory(defines.inventory.assembling_machine_input).insert( {name = "mind-controller-nothing", count = 1000} )
				controller.entity_visible = ent
				controller.entity_invisible.recipe = "mind-controller-charge"
				controller.inv_input.insert( {name = "mind-controller-charge-empty", count = 1} )
				controller.inv_output.clear()
				
			elseif controller.status == status_charging then
				local nb_charge_full = controller.inv_output.get_item_count("mind-controller-charge-full")
				
				if nb_charge_full >= 1 then
					debug_print( "scanning!" )
					controller.status = status_scanning
					
					ent = controller.entity_visible
					if ent then ent.destroy() end
					-- ent = controller.surface.create_entity({name = "mind-controller-scanning", force = game.forces.neutral, position = pos_offset(controller.position,{x=2,y=0})})
					ent = controller.surface.create_entity({name = "mind-controller-scanning", force = game.forces.neutral, position = controller.position})
					ent.operable = false
					ent.get_inventory(defines.inventory.assembling_machine_input).insert( {name = "mind-controller-nothing", count = 1000} )
					controller.entity_visible = ent
				end
			
			elseif controller.status == status_scanning then
				local enns = controller.surface.find_enemy_units( controller.position, mind_controller_range, controller.force )
				
				if #enns > 0 then
					-- try to mind control some aliens
										
					local nb, nbmax
					nb = 0
					nbmax = mind_controller_max_control_per_cycle + math.random(-1,1)
					
					for _, enn in pairs(enns) do
						if is_controlable(enn) then
							local enn_name = "controled-" .. enn.name
							local enn_health = enn.health
							local enn_position = enn.position
							enn.destroy()
							local friend = controller.surface.create_entity({name = enn_name, force = controller.force, position = enn_position})
							friend.health = enn_health
							table.insert(controller.friends, friend)
							controller.surface.create_entity({name = "flying-text", position = enn_position, text = "@", color = {r=1, g=0, b=1}})
							nb=nb+1
							nbmax = nbmax - 1
							if nbmax <= 0 then
								break
							end
						end
					end

					if nb > 0 then
						debug_print( "starting emission!" )				
						controller.status = status_emitting
						
						ent = controller.entity_visible
						if ent then ent.destroy() end
						-- ent = controller.surface.create_entity({name = "mind-controller-emitting", force = game.forces.neutral, position = pos_offset(controller.position,{x=2,y=0})})
						ent = controller.surface.create_entity({name = "mind-controller-emitting", force = game.forces.neutral, position = controller.position})
						ent.operable = false
						ent.get_inventory(defines.inventory.assembling_machine_input).insert( {name = "mind-controller-nothing", count = 1000} )
						controller.entity_visible = ent
						controller.entity_invisible.recipe = "mind-controller-emission"
						controller.inv_input.insert( {name = "mind-controller-charge-full", count = 1} )
						controller.inv_output.clear()
					end
				end
				
			elseif controller.status == status_emitting then
				local nb_emissions = controller.inv_output.get_item_count("mind-controller-emission")
				
				if nb_emissions >= 1 then
					debug_print( "ending emission!" )
					controller.status = status_created
				end
			end
		end
			
	elseif global.tick == 38 then
		for _, controller in pairs(global.controllers) do	
			-- reduce aliens health

			local i = 1

			while i <= #controller.friends do
				local friend = controller.friends[i]
				
				if friend and friend.valid then
					friend.health = friend.health - mind_controller_reduce_health * friend.prototype.max_health
				
					if friend.health <= 0 then
						controller.surface.create_entity({name = "flying-text", position = friend.position, text = "+", color = {r=1, g=0, b=1}})
						friend.die()
						table.remove(controller.friends,i)
						i = i-1
					end
				else
					table.remove(controller.friends,i)
					i = i-1
				end
				
				i=i+1
			end
			
			-- release some aliens and clean list

			if math.random() > 1 - mind_controller_release_probability then
				local enn, nb
				i = 1
				nb = math.floor( #controller.friends * mind_controller_release_rate + 0.5 )

				while i <= #controller.friends do
					friend = controller.friends[i]
					
					if friend and friend.valid then
						if i <= nb then
							release_friend(controller,friend)
							table.remove(controller.friends,i)
							nb = nb - 1
							i = i-1
						end
					else
						table.remove(controller.friends,i)
						i = i-1
					end
					
					i=i+1
				end
			end
		end
		
	elseif global.tick == 19 then
		-- manage released aliens that are uncontrolable for a few seconds
		local enn, i
		i = 1
		while i <= #global.released do
			enn = global.released[i]
			if enn.entity and enn.entity.valid then
				if enn.tick <= game.tick then
					enn = nil
				end
			else
				enn = nil
			end
			
			if enn == nil then
				table.remove( global.released, i )
				i = i-1
			end
			i=i+1
		end
			
		if game.tick % 300 < 60 then
			debug_print( "rel=", #global.released )
		end
	end
	
	global.tick = global.tick - 1
end

script.on_event(defines.events.on_tick, on_tick )

--------------------------------------------------------------------------------------
function release_friend(controller,friend)
	local enn_name = string.sub(friend.name, 11 )
	local enn_health = friend.health
	local enn_position = friend.position
	friend.destroy()
	local enn = controller.surface.create_entity({name = enn_name, force = game.forces.enemy, position = enn_position})
	enn.health = enn_health
	table.insert( global.released,
		{ entity = enn, tick = game.tick + mind_controller_release_time * 60 }
	)
	controller.surface.create_entity({name = "flying-text", position = enn_position, text = "x", color = {r=1, g=0, b=0}})
end

--------------------------------------------------------------------------------------
function is_controlable(enn)
	for _, rel in pairs( global.released ) do
		if rel.entity == enn then 
			return( false )
		end
	end
	
	return( true )
end

--------------------------------------------------------------------------------------

local interface = {}

function interface.reset()
	debug_print( "reset" )
	
	global.tick = 0
	global.controllers = {}
	global.released = {}
	
	for _,force in pairs(game.forces) do
		force.reset_recipes()
		force.reset_technologies()
	end
end

remote.add_interface( "megamind", interface )

-- /c remote.call( "megamind", "reset" )